/*
 *	File    : pthreads.c
 *	Title   : Demo Producer/Consumer
 *	Desc    : Sample solution to the producer/consumer problem using pthreads.
 *	Author  : Andrae Muys
 *	Date    : 18 September 1997
 *	Revised	: 20 April 2015, Mike Lam (JMU)
 */

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>         /* usleep */

#define QUEUESIZE 5
#define LOOP 20

typedef int data;

typedef struct {
	data buffer[QUEUESIZE];
	int front, back, len;
	pthread_mutex_t *lock;
	pthread_cond_t *not_full, *not_empty;
} queue;

queue *q_alloc(void)
{
	queue *q;

    /* allocate buffer for data */
	q = (queue *)malloc (sizeof (queue));
	if (q == NULL) return (NULL);
	q->front = q->back = q->len = 0;

    /* allocate mutex and wait conditions */
	q->lock = (pthread_mutex_t *) malloc(sizeof (pthread_mutex_t));
	pthread_mutex_init(q->lock, NULL);
	q->not_full = (pthread_cond_t *) malloc(sizeof (pthread_cond_t));
	pthread_cond_init(q->not_full, NULL);
	q->not_empty = (pthread_cond_t *) malloc(sizeof (pthread_cond_t));
	pthread_cond_init(q->not_empty, NULL);

	return (q);
}

void q_free (queue *q)
{
    /* clean up */
	pthread_mutex_destroy(q->lock);
	free(q->lock);
	pthread_cond_destroy(q->not_full);
	free(q->not_full);
	pthread_cond_destroy(q->not_empty);
	free(q->not_empty);
	free(q);
}

void q_enqueue(queue *q, data in)
{
    /* acquire lock on the queue */
    pthread_mutex_lock(q->lock);

    /* if the queue is already full, wait until it's not full */
    while (q->len == QUEUESIZE) {
        pthread_cond_wait(q->not_full, q->lock);
    }

    /* insert into queue */
	q->buffer[q->back] = in;
	q->back = (q->back+1) % QUEUESIZE;
    q->len++;
    printf("ENQUEUE: %d\n", in);

    /* release lock on the queue */
    pthread_mutex_unlock (q->lock);

    /* signal anything that was waiting b/c the queue was empty */
    pthread_cond_signal(q->not_empty);

	return;
}

void q_dequeue(queue *q, data *out)
{
    /* basically the opposite of enqueue */

    pthread_mutex_lock(q->lock);
    while (q->len == 0) {
        pthread_cond_wait(q->not_empty, q->lock);
    }
	*out = q->buffer[q->front];
	q->front = (q->front+1) % QUEUESIZE;
    q->len--;
    printf("DEQUEUE: %d\n", *out);
    pthread_mutex_unlock(q->lock);
    pthread_cond_signal(q->not_full);
	return;
}

void *seq_producer(void *q)
{
	queue *fifo;
	data i;

	fifo = (queue *)q;
	for (i = 0; i < LOOP; i++) {
		q_enqueue(q, i);
		usleep (100000);
	}
	return (NULL);
}

void *print_consumer(void *q)
{
	queue *fifo;
    data d;
	int i;

	fifo = (queue *)q;
	for (i = 0; i < LOOP; i++) {
		q_dequeue (fifo, &d);
		printf ("%d\n", d);
		usleep(200000);
	}
	return (NULL);
}

int main ()
{
	queue *fifo;
	pthread_t pro, con;

    /* allocate queue */
	fifo = q_alloc();
	if (fifo ==  NULL) {
		fprintf (stderr, "Out of memory!\n");
		exit (1);
	}

    /* fork and join threads */
	pthread_create(&pro, NULL, seq_producer, fifo);
	pthread_create(&con, NULL, print_consumer, fifo);
	pthread_join(pro, NULL);
	pthread_join(con, NULL);

    /* clean up queue */
	q_free(fifo);

	return 0;
}

